#include <avr/io.h>
#include "device.h"
#include "ir_rc5.h"

/* RC5 Protocol Decoding */

/* The formula to calculate ticks is as follows
 * TICKS = PULSE_LENGTH * (CPU_FREQ / TIMER_PRESCALER)
 * Where CPU_FREQ is given in MHz and PULSE_LENGTH in us.
 * LONG_MIN should usually be SHORT_MAX + 1 */
#define SHORT_MIN 444   /* 444 microseconds */
#define SHORT_MAX 1333  /* 1333 microseconds */
#define LONG_MIN 1334   /* 1334 microseconds */
#define LONG_MAX 2222   /* 2222 microseconds */

typedef enum {
    START1 = 0,
    MID1 = 1,
    MID0 = 2,
    START0 = 3,
    BEGIN = 4
} RC5_STATE;

typedef enum {
    SHORTSPACE = 0,
    SHORTPULSE = 2,
    LONGSPACE  = 4,
    LONGPULSE  = 6,
    ERROR      = 8
} RC5_EVENT;

#define LONG(event) event + 4

/* trans[] is a table of transitions, indexed by
 * the current state.  Each byte in the table
 * represents a set of 4 possible next states,
 * packed as 4 x 2-bit values: 8 bits DDCCBBAA,
 * where AA are the low two bits, and
 *   AA = short space transition
 *   BB = short pulse transition
 *   CC = long space transition
 *   DD = long pulse transition
 *
 * If a transition does not change the state,
 * an error has occured and the state machine should
 * reset.
 *
 * The transition table is:
 * 00 00 00 01  from state 0: short space->1
 * 10 01 00 01  from state 1: short pulse->0, long pulse->2
 * 10 01 10 11  from state 2: short space->3, long space->1
 * 11 11 10 11  from state 3: short pulse->2
 */
static const uint8_t trans[4] = {0x01, 0x91, 0x9b, 0xfb};
static uint16_t rc5, data;
static uint8_t counter;
static RC5_STATE state = BEGIN;
/* Emit 0 - just decrement bit position counter
* cause data is already zeroed by default. */
#define EMMIT_0 counter--
#define EMMIT_1 counter--; rc5 |= (1<<counter)
#define START_BITS(data) ((data) & 0x3000)>>12)
#define TOGGLE_BIT(data) ((data) & 0x0800)>>11
#define ADDRESS(data) ((data) & 0x07C0)>>6
#define COMMAND(data) (data) & 0x003F

void RC5_Reset(void) {
    counter = 14;
    rc5 = 0;
    state = BEGIN;
}

uint8_t RC5_Address(void) {
    return ((data) & 0x07C0)>>6;
}

uint8_t RC5_Command(void) {
    return (data) & 0x003F;
}

uint8_t RC5_Decode(uint16_t delay) {
    //
    if(state == BEGIN) {
        EMMIT_1;
        state = MID1;
        return 0;
    }

    /* TSOP2236 pulls the data line up, giving active low,
     * so the output is inverted. If data pin is high then the edge
     * was falling and vice versa.
     */
    RC5_EVENT event = (PINB & IR) ? SHORTPULSE : SHORTSPACE;

    if(delay > LONG_MIN && delay < LONG_MAX) {
        event = LONG(event);
    } else if(delay < SHORT_MIN || delay > SHORT_MAX) {
        /* If delay wasn't long and isn't short then
         * it is erroneous so we need to reset but
         * we don't return from interrupt so we don't
         * loose the edge currently detected. */
        RC5_Reset();
    }

    RC5_STATE newstate = (trans[state] >> event) & 0x03;

    if(newstate == state || state > START0) {
        /* No state change or wrong state means
         * error so reset. */
        RC5_Reset();
        return 0;
    }

    state = newstate;

    if(state == MID0) {
        EMMIT_0;
    } else if(state == MID1) {
        EMMIT_1;
    }

    /* The only valid end states are MID0 and START1.
     * Mid0 is ok, but if we finish in MID1 we need to wait
     * for START1 so the last edge is consumed. */
    if(counter == 0 && (state == START1 || state == MID0)) {
        data = rc5;
        RC5_Reset();
        return 1;
    }
    return 0;
}
